/*
 * CLIF is a Load Injection Framework
 * Copyright (C) 2006 F.Barreca & I.Vassilev
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * CLIF
 *
 * Contact: clif@objectweb.org
 */

package org.objectweb.clif.isac.plugins.sipinjector10;

// System imports
import java.util.*;

// CLIF imports
import org.objectweb.clif.scenario.isac.plugin.SampleAction;
import org.objectweb.clif.scenario.isac.plugin.TestAction;
import org.objectweb.clif.scenario.isac.plugin.ControlAction;
import org.objectweb.clif.scenario.isac.plugin.SessionObjectAction;
import org.objectweb.clif.storage.api.ActionEvent;
import org.objectweb.clif.scenario.isac.exception.IsacRuntimeException;


// SIP imports
import javax.sip.*;
import javax.sip.address.*;
import javax.sip.header.*;
import javax.sip.message.*;

/**
 * @author Fabien Barreca
 * @author Iassen Vassilev
 */
public class SessionObject implements SampleAction, TestAction, ControlAction,
		SessionObjectAction, SipListener {

	// Session object's attributes - SIP
	private String transportType;
	private String serverIPAddress;
	private String serverSIPUser;
	private String serverSIPDomain;
	private String myIPAddress;
	private String myUserName;
	
	// Session object's attributes - SDP
	private String SdpV;
	private String SdpO;
	private String SdpS;
	private String SdpC;
	private String SdpT;
	private String SdpM;
	private ArrayList SdpA;
	
	// Attributes related to the SIP Communications
	private SipStack sipStack;
	private SipProvider sipProvider;
	private AddressFactory addressFactory;
	private HeaderFactory headerFactory;
	private MessageFactory messageFactory;
	private ContactHeader contactHeader;
	private ListeningPoint listeningPoint;
	private Dialog dialog;
	protected ClientTransaction firstTID;
	protected ServerTransaction ServerTID;
	private Request lastRequest;
	private SessionObject myMaster;
	private long dateInvite;
	private long timeOut;
	public int lastResponseCode;
	private int lastRequestCode;
	private boolean isWaitingForRequest;
	private String lastControlRequestSent;
	private int requestSeqNbr;
	private int myPort;
	private String serverPort;

	// Clone objects management
	private static int ID_TAG_CNT;
	private static int ACTIVE_CLONES;
	public int myIdTag;
	private ArrayList listObjects;
	private ArrayList listResponseCode;
	
	
	/**
	 * Constructor
	 * 
	 * @param params : defining the SIPInjector SessionObject
	 */
	public SessionObject(Hashtable params) {

		// in case of trying to RE-initialise : must shutDown before
		if (sipStack != null)
			shutDown();
		
		// Getting the SessionObject basic attributes
		transportType = (String) params.get(Constants.STR_PARAM_TRANSPORTTYPE);
		serverSIPUser = (String) params.get(Constants.STR_PARAM_SERVERSIPUSER);
		serverSIPDomain = (String) params.get(Constants.STR_PARAM_SERVERSIPDOMAIN);
		serverIPAddress = (String) params.get(Constants.STR_PARAM_SERVERIPADDRESS);
		myIPAddress = (String) params.get(Constants.STR_PARAM_MYIPADDRESS);
		myUserName = (String) params.get(Constants.STR_PARAM_MYUSERNAME);
		myUserName = (String) params.get(Constants.STR_PARAM_MYUSERNAME); 
		
		// Getting the timeOut value if set by user else use default
		String strTimeOut = (String) params.get(Constants.STR_PARAM_TIMEOUT);
		if (strTimeOut.equals(""))
			timeOut = Constants.DEFAULT_TIMEOUT;
		else
			timeOut = Long.parseLong(strTimeOut);
		
		// Getting the ports values if set by user else use default
		String strMyPort = (String) params.get(Constants.STR_PARAM_MYLISTENINGPORT);
		if (strMyPort.equals(""))
			myPort = Constants.DEFAULT_MYPORT;
		else
			myPort = Integer.parseInt(strMyPort);
		
		String strServerPort = (String) params.get(Constants.STR_PARAM_SERVERLISTENINGPORT);
		if (strServerPort.equals(""))
			serverPort = Constants.DEFAULT_STR_SERVERPORT;
		else
			serverPort = strServerPort;
		
		// Initialising the object fields
		lastResponseCode = -1;
		lastRequestCode = -1;
		isWaitingForRequest = false;
		dialog = null;
		listObjects = new ArrayList();
		listResponseCode = new ArrayList();

		// Getting the SessionObject basic attributes related to SDP
		SdpV="v="+(String) params.get(Constants.STR_PARAM_SDP_V)+"\r\n";
		SdpO="o="+(String) params.get(Constants.STR_PARAM_SDP_O)+"\r\n";
		SdpS="s="+(String) params.get(Constants.STR_PARAM_SDP_S)+"\r\n";
		SdpC="c="+(String) params.get(Constants.STR_PARAM_SDP_C)+"\r\n";
		SdpT="t="+(String) params.get(Constants.STR_PARAM_SDP_T)+"\r\n";
		SdpM="m="+(String) params.get(Constants.STR_PARAM_SDP_M)+"\r\n";
		SdpA=new ArrayList();
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A1)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A1)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A2)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A2)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A3)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A3)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A4)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A4)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A5)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A5)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A6)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A6)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A7)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A7)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A8)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A8)+"\r\n");
		if (!(((String)params.get(Constants.STR_PARAM_SDP_A9)).equals("")))
			SdpA.add("a="+(String)params.get(Constants.STR_PARAM_SDP_A9)+"\r\n");
		
		// Initializing the SIP Stack
		initSipStack();
		
		// Initialising the clones counters 
		ID_TAG_CNT = 0;
		ACTIVE_CLONES = 0;
		requestSeqNbr = 0;

	}

	/**
	 * Constructor used in cloneObject method
	 * 
	 * @param toClone : The specimen object to clone
	 */
	private SessionObject(SessionObject toClone) {

		// Getting the SessionObject basic attributes
		transportType = toClone.getTransportType();
		serverSIPUser = toClone.getServerSIPUser();
		serverSIPDomain = toClone.getServerSIPDomain();
		serverIPAddress = toClone.getServerIPAddress();
		myIPAddress = toClone.getMyIPAddress();
		myUserName = toClone.getMyUserName();
		lastResponseCode = -1;
		lastRequestCode = -1;
		myPort = toClone.getMyPort();
		serverPort = toClone.getServerPort();
		isWaitingForRequest = false;
		dialog = toClone.getDialog();
		
		// Getting the SessionObject basic attributes related to SDP
		SdpV=toClone.SdpV;
		SdpO=toClone.SdpO;
		SdpS=toClone.SdpS;
		SdpC=toClone.SdpC;
		SdpT=toClone.SdpT;
		SdpM=toClone.SdpM;
		SdpA=toClone.SdpA;
		
		// Getting a reference to the SipStack objects
		sipStack = toClone.sipStack;
		sipProvider= toClone.sipProvider;
		addressFactory= toClone.addressFactory;
		headerFactory=toClone.headerFactory;
		messageFactory=toClone.messageFactory;
		contactHeader= toClone.contactHeader;
		listeningPoint = toClone.listeningPoint;
		timeOut = toClone.getTimeout();
		
		// Management of the clones
		ID_TAG_CNT = ID_TAG_CNT + 1;
		ACTIVE_CLONES = ACTIVE_CLONES + 1;
		myIdTag = ID_TAG_CNT;
		toClone.listObjects.add(this);
		this.myMaster = toClone;
		Semaphore s1=new Semaphore(0,timeOut);
		toClone.listResponseCode.add(s1);
		requestSeqNbr = 0;
	}

	/***************************************************************************
	 * SESSIONOBJECTACTION INTERFACE IMPLEMENTATION *
	 **************************************************************************/
	
	/**
	 * Method used by the specimen object to create new clones
	 * 
	 * @param toClone : The specimen object to clone
	 */
	public Object createNewSessionObject() {
		return new SessionObject(this);
	}

	/**
	 * Method called by each clone at the end of 
	 * the execution of each ISAC behavior tree
	 */
	public void reset() {
	}

	/**
	 * Method called by each clone at the end of 
	 * the execution of an ISAC scenario. 
	 * Tests if the clone is the last active : in this case it shuts Down
	 * the SipStack.
	 */
	public void close() {
		ACTIVE_CLONES=ACTIVE_CLONES - 1;
		if (ACTIVE_CLONES==0)
			myMaster.shutDown();
	}

	/***************************************************************************
	 * SAMPLEACTION INTERFACE IMPLEMENTATION *
	 **************************************************************************/

	/**
	 * This method executes a sample action. Calls doRequest() to do the job,
	 * calculates the duration of the request and assignes the duration in the
	 * report object.
	 * 
	 * @param number :
	 *            The number of the action to execute
	 * @param params : 
	 *            A map of the parameters of the action
	 * @param report : 
	 *            Will contain the result of the accomplished action
	 * 
	 */
	public ActionEvent doSample(int number, Map params, ActionEvent report) {
		
		ActionEvent res;
		report.setDate(System.currentTimeMillis());
		res = doRequest(number, report);
		if (number != Constants.SAMPLE_WAIT)
			res.duration = (int) (System.currentTimeMillis() - res.getDate());
		return res;
	}

	/***************************************************************************
	 * TESTACTION INTERFACE IMPLEMENTATION *
	 **************************************************************************/

	/**
	 * This method will execute a test (dispatcher). Tests the lastResponseCode and
	 * the lastRequestCode 
	 * 
	 * @param number :
	 *            The number of the test to execute
	 * @param params :
	 *            A map of the parameters of the test
	 * 
	 */
	public boolean doTest(int number, Map params) {

		// Dispatch to the test to execute
		switch (number) {
		case Constants.TEST_RESPONSEISCODE:
			int statusCode = new Integer((String) params.get(Constants.STR_TEST_PARAM_RESPONSEISCODE)).intValue();
			return (lastResponseCode == statusCode);
		case Constants.TEST_RESPONSEIS1XX:
			return((lastResponseCode >= 100) && (lastResponseCode < 200));
		case Constants.TEST_RESPONSEIS2XX:
			return((lastResponseCode >= 200) && (lastResponseCode < 300));
		case Constants.TEST_RESPONSEIS3XX:
			return((lastResponseCode >= 300) && (lastResponseCode < 400));
		case Constants.TEST_RESPONSEIS4XX:
			return((lastResponseCode >= 400) && (lastResponseCode < 500));
		case Constants.TEST_RESPONSEIS5XX:
			return((lastResponseCode >= 500) && (lastResponseCode < 600));
		case Constants.TEST_RESPONSEIS6XX:
			return((lastResponseCode >= 600) && (lastResponseCode < 700));
		case Constants.TEST_REQUESTISINVITE:
			return(lastRequestCode == Constants.CONTROL_INVITE);
		case Constants.TEST_REQUESTISACK:
			return(lastRequestCode == Constants.CONTROL_ACK);
		case Constants.TEST_REQUESTISCANCEL:
			return(lastRequestCode == Constants.SAMPLE_CANCEL);
		case Constants.TEST_REQUESTISBYE:
			return(lastRequestCode == Constants.SAMPLE_BYE);
		default:
		}
		return false;
	}

	/***************************************************************************
	 * CONTROLACTION INTERFACE IMPLEMENTATION *
	 **************************************************************************/
	/**
	 * This method will execute a control action. 
	 * 
	 * @param number :
	 *            The number of the control action to execute
	 * @param params :
	 *            A map of the parameters of the control action
	 * 
	 */
public void doControl(int number, Map params) {

		// Dispatch to the control action to execute
		//setting the primitive to execute
		switch (number) {
		case Constants.CONTROL_RESPOND:
			doControlRespond(params);
			break;
		case Constants.CONTROL_WAITREQUEST:
			doWaitRequest();
			break;
		default :
			doControlRequest(number);
		}
}

	
	/***************************************************************************
	 * SIPLISTENER INTERFACE IMPLEMENTATION *
	 **************************************************************************/
	/**
	 * Event-based method called by the SIP stack when a request is received
	 * The "master" object gives the request to the targeted SessionObject
	 * and sets it free from the semaphore by a call to V()
	 * Used only by the "master" object
	 * 
	 * @param requestEvent :
	 *            the received request object
	 */
	public void processRequest(RequestEvent requestEvent) {
		// Getting the response tag
		ServerTransaction st = requestEvent.getServerTransaction();
		String tagResponseStr = st.getDialog().getLocalTag();
		int tagResponseInt = Integer.parseInt(tagResponseStr); 
			
		SessionObject mo;
		Semaphore so;
		int requestNbr;
		try {
			
			String strMethod = requestEvent.getRequest().getMethod();
			// getting the sessionObject which is targeted by the received request
			mo = (SessionObject)listObjects.get(tagResponseInt - 1);
			
			// if the object is not waiting for a request...skip the request
			if (!mo.isWaitingForRequest)
				return;
			
			// Getting the rigth semaphore object
			so=(Semaphore) listResponseCode.get(tagResponseInt-1);
			
			//setting the primitive to execute
			if (strMethod.equals(Request.CANCEL))
				requestNbr = Constants.SAMPLE_CANCEL;
			else
				if (strMethod.equals(Request.ACK))
					requestNbr = Constants.CONTROL_ACK;
				else
					if (strMethod.equals(Request.BYE))
						requestNbr = Constants.SAMPLE_BYE;
					else
						if (strMethod.equals(Request.INVITE))
							requestNbr = Constants.CONTROL_INVITE;
						else
							requestNbr = Constants.CONTROL_ACK;
			
			// setting the object's dialog and transaction attributes
			mo.dialog = st.getDialog();
			mo.ServerTID = st;
			mo.lastRequest = requestEvent.getRequest();
			
			// frees the targeted object from the semaphore with the code requestNbr
			so.V(requestNbr);
			
			// sets the lastrequestCode for the tests
			mo.lastRequestCode = requestNbr;
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
 	 * Event-based method called by the SIP stack when a response is received
	 * The "master" object gives the response to the targeted SessionObject
	 * and sets it free from the semaphore by a call to V()
	 * Used only by the "master" object
	 * 
	 * @param responseEvent :
	 *            the object received in response
	 */
public void processResponse(ResponseEvent responseEvent) {
		// Getting the response tag to know to wich of the object the response is targeted
		ClientTransaction ct = responseEvent.getClientTransaction();
		String tagResponseStr;
		if (ct.getDialog() !=  null)
			tagResponseStr = ct.getDialog().getLocalTag();
		else {
			String header = ct.getRequest().getHeader(Constants.STR_HEADERNAME).toString();
			tagResponseStr = getTagFromIdHeader(header);
		}	
		
		
		SessionObject mo;
		Semaphore so;
		try {
			int tagResponseInt = Integer.parseInt(tagResponseStr);
			
			// getting the targeted object and the corresponding semphore
			so=(Semaphore) listResponseCode.get(tagResponseInt-1);
			mo = (SessionObject)listObjects.get(tagResponseInt - 1);
			
			String seqHeader = ct.getRequest().getHeader(Constants.STR_HEADERCSEQ).toString();
			String seqNbrStr = getTagFromIdHeader(seqHeader);
			int seqNbrInt = Integer.parseInt(seqNbrStr);
			
			// if the response comes too late...ignore it
			if (seqNbrInt < mo.requestSeqNbr){
				return;
			}
			// set the object attributes related to the dialog
			mo.dialog = ct.getDialog();
			mo.firstTID = ct;	
			
			// freeing the object from the semaphore
			so.V(responseEvent.getResponse().getStatusCode());
		} catch (Exception e) {
			e.printStackTrace();
		}		
	}



/**
 * Event-based method called by the SIP stack when a received request is
 * the state of "timeout" (is not handled)
 * 
 * Used only by the "master" object
 * 
 * @param responseEvent :
 *            the object received in response
 */
	public void processTimeout(TimeoutEvent timeoutEvent) {
		
		ClientTransaction ct =timeoutEvent.getClientTransaction();
		if(ct!=null){
			String tagResponseStr = ct.getDialog().getLocalTag();
			int tagResponseInt = Integer.parseInt(tagResponseStr); 
				
			SessionObject mo;
			//Semaphore so;
			try {
				//so=(Semaphore) listResponseCode.get(tagResponseInt-1);
				//so.V(33);
				
				mo = (SessionObject)listObjects.get(tagResponseInt - 1);
				mo.dialog = ct.getDialog();
				mo.firstTID = ct;
			} catch (Exception e) {
				e.printStackTrace();
			}			
		}
	}


	/***************************************************************************
	 * SIP STACK MANAGEMENT (INIT...SHUTDOWN...)*
	 **************************************************************************/

	/**
	 * 
	 * Initializing the SIP Stack - core processing and setting all the common
	 * fields of the SessionObject related to the SIP dialog. Setting anf initialising the headers
	 * 
	 */
	public void initSipStack() {
		// if a sipstack is already initialised...must shut down
		if(sipStack != null)
		{
			shutDown();
		}
		
		// getting the singleton sipFactory
		SipFactory sipFactory = null;
		sipStack = null;
		sipFactory = SipFactory.getInstance();
		sipFactory.setPathName(Constants.STR_SIP_IMPLEMENTATION);
		
		// setting the properties to be applied to the sipStack
		Properties properties = new Properties();

		String peerHostPort = serverIPAddress + ":"
				+ serverPort;

		properties.setProperty("javax.sip.OUTBOUND_PROXY", peerHostPort + "/"
				+ transportType);

		if (transportType.equals(Constants.STR_TRANSPORTTYPE_UDP))
			properties.setProperty("javax.sip.ROUTER_PATH",
					"org.objectweb.clif.isac.plugins.sipinjector10.MyRouter");

		properties.setProperty("javax.sip.IP_ADDRESS", myIPAddress);
		properties.setProperty("javax.sip.STACK_NAME", "sipinjector10");
		properties.setProperty("javax.sip.RETRANSMISSION_FILTER", "on");



		// Drop the client connection after we are done with the transaction.
		properties.setProperty("gov.nist.javax.sip.CACHE_CLIENT_CONNECTIONS",
				"false");

		// Set to 0 in your production code for max speed.
		// You need 16 for logging traces. 32 for debug + traces.
		// Your code will limp at 32 but it is best for debugging.
		//properties.setProperty("gov.nist.javax.sip.TRACE_LEVEL", "32");

		try {
			// Create SipStack object with the given properties
			sipStack = sipFactory.createSipStack(properties);
		} catch (Exception e) {
			throw new IsacRuntimeException("SipStack Init Failed : " + e);
		}

		try {
			// Initialising the factories to be used forconstructing the queries
			headerFactory = sipFactory.createHeaderFactory();
			addressFactory = sipFactory.createAddressFactory();
			messageFactory = sipFactory.createMessageFactory();
			listeningPoint = sipStack.createListeningPoint(
					myPort, transportType);

			sipProvider = sipStack.createSipProvider(listeningPoint);
			SessionObject listener = this;
			sipProvider.addSipListener(listener);

		} catch (Exception e) {
			// System.out.println(ex.getMessage());
			throw new IsacRuntimeException("Factories creation failed : " + e);
		}

	}
	
	/**
	 * Shutting down the sipStack and sets to null all the Sip
	 * dialog related fields
	 * 
	 */
	public void shutDown() {
		try{
			sipStack.deleteListeningPoint(listeningPoint);
			sipProvider.removeSipListener(this);
			while(true){
				try {
					sipStack.deleteSipProvider(sipProvider);
					break;
				} catch (ObjectInUseException ex1){
						continue;
					}
			}	
			
			sipStack=null;
			sipProvider=null;
			firstTID=null;
			contactHeader=null;
			addressFactory=null;
			headerFactory=null;
			messageFactory=null;
			listeningPoint=null;
			SipFactory sipFactory = null;
			sipFactory = SipFactory.getInstance();
			sipFactory.resetFactory();
		}catch (Exception ex){
			throw new IsacRuntimeException("SipStack destruction failed : " + ex);
		}
	
	}

/***************************************************************************
 * METHODS WHICH SEND SIP REQUESTS*
 **************************************************************************/
	/**
	 * Constructs and sends a "regular" SIP request. Constructs the headers,
	 * sends the request, waits on a semaphore for the response and sets the lastResponseCode
	 * field. In the case of a WAIT request does not send anything, just waits for a response
	 * to come and sets the lastResponseCode
	 * @param
	 * 		number : the "request to do" number
	 * @param
	 * 		report : the actionReport to fill
	 * @return
	 * 		the actionEvent report with the response code and the duration filled
	 * 
	 */	
public ActionEvent doRequest(int number, ActionEvent report)
	{
		// Getting the name of the last ControlRequest sent
		String strRequest= lastControlRequestSent;
		// re-initialises the lastResponseCode
		lastResponseCode = -1;
		try {
			if(number!=Constants.SAMPLE_WAIT)
			{
				requestSeqNbr = requestSeqNbr + 1;
				// create >From Header
				SipURI fromAddress = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, Constants.STR_SIP_FROMADDRESS);
				Address fromNameAddress = addressFactory.createAddress(fromAddress);
				fromNameAddress.setDisplayName(Constants.STR_SIP_FROMDISPLAYNAME);
				FromHeader fromHeader = headerFactory.createFromHeader(fromNameAddress, String.valueOf(myIdTag));
			
				// create To Header
				SipURI toAddress = addressFactory.createSipURI(serverSIPUser, serverSIPDomain);
				Address toNameAddress = addressFactory.createAddress(toAddress);
				toNameAddress.setDisplayName(Constants.STR_SIP_TODISPLAYNAME);
				ToHeader toHeader = headerFactory.createToHeader(toNameAddress, null);
			
				// create Request URI
				String addressAndPort = serverIPAddress + ":" + serverPort; 
				SipURI requestURI = addressFactory.createSipURI(serverSIPUser, addressAndPort);
			
				// Create ViaHeaders
				ArrayList viaHeaders = new ArrayList();
				ViaHeader viaHeader = headerFactory.createViaHeader(
					sipStack.getIPAddress(),
					sipProvider.getListeningPoint().getPort(),
					transportType,
					null);
			
				// add via headers
				viaHeaders.add(viaHeader);

				// Create ContentTypeHeader
				ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("application", "sdp");

				// Create a new CallId header
				CallIdHeader callIdHeader = sipProvider.getNewCallId();

				// Create a new MaxForwardsHeader
				MaxForwardsHeader maxForwards = headerFactory.createMaxForwardsHeader(70);

				// setting the primitive to execute
				switch (number) {
				case Constants.SAMPLE_CANCEL:
					strRequest = Request.CANCEL;
					break;
				case Constants.SAMPLE_BYE:
					strRequest = Request.BYE;
					break;
				default :
					strRequest = Request.BYE;
			}
			
				// Create a new Cseq header
				CSeqHeader cSeqHeader =
				headerFactory.createCSeqHeader(1, strRequest);
			
				// Create the request.
				Request request = messageFactory.createRequest(
					requestURI,
					strRequest,
					callIdHeader,
					cSeqHeader,
					fromHeader,
					toHeader,
					viaHeaders,
					maxForwards);
	
				String host = sipStack.getIPAddress();

				SipURI contactUrl = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, host);
				contactUrl.setPort(listeningPoint.getPort());
			
				// Create the contact name address.
				SipURI contactURI = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, host);
				contactURI.setPort(sipProvider.getListeningPoint().getPort());

				Address contactAddress = addressFactory.createAddress(contactURI);

				// Add the contact address.
				contactAddress.setDisplayName(Constants.STR_SIP_FROMNAME);

				contactHeader = headerFactory.createContactHeader(contactAddress);
				request.addHeader(contactHeader);

				// Add the extension header for the SessionObject ID 
				Header extensionHeader =
				headerFactory.createHeader(Constants.STR_HEADERNAME, String.valueOf(myIdTag));
				request.addHeader(extensionHeader);
				
				//Add the extension header for the sequence ID
				Header extensionHeader2 =
					headerFactory.createHeader(Constants.STR_HEADERCSEQ, String.valueOf(requestSeqNbr));
				request.addHeader(extensionHeader2);
	

				// set the parameters for the SDP
				String sdpData = SdpV
					+ SdpO
					+ SdpS
					+ SdpC
					+ SdpT
					+ SdpM;
				for (int i=0; i<SdpA.size();i++)
					sdpData=sdpData+SdpA.get(i);
		
				request.setContent(sdpData, contentTypeHeader);

				Header callInfoHeader =
				headerFactory.createHeader(
					"Call-Info",
					"<http://www.antd.nist.gov>");
				request.addHeader(callInfoHeader);


				firstTID = sipProvider.getNewClientTransaction(request);
				firstTID.sendRequest();
			}
			
			// getting the semaphore and waiting for the response
			Semaphore so = (Semaphore)myMaster.listResponseCode.get((myIdTag-1)); 
			so.P();
			lastResponseCode = so.getValeur();
			
			// In case of wait sets the duration beginning the the wright time
			if (number==Constants.SAMPLE_WAIT)
			{
				report.duration = (int)(System.currentTimeMillis()- dateInvite); 
			}
			
			//	In case of timeOut...the action is concidered to be an error
			if (lastResponseCode == 0){
				report.successful = false;
				report.comment = "Request : " + strRequest + " sent";
				report.result= "SIP RESPONSE: "+ String.valueOf(lastResponseCode) +" ID TAG: " + myIdTag;	
			}
			else {
				report.successful = true;
				report.comment = "Request : " + strRequest + " sent";
				report.result= "SIP RESPONSE: "+ String.valueOf(lastResponseCode) +" ID TAG: " + myIdTag;
			}
			return report;
			
		}catch(Exception e) {
			report.successful = false;
			report.comment = "exception thrown";
			throw new IsacRuntimeException("DoRequest failed : " + e);
		}
		
	}	

/**
 * Constructs and sends a "control" request. Constructs the headers and
 * sends the request without waiting for a response.
 * @param
 * 		number : the "request to do" number
 * 
 */
public void doControlRequest (int number)
{
	String strRequest;
	String user;
	
	// we need to identify ourself to the server we want to register to
	if(number ==Constants.CONTROL_REGISTER)
	{
		user = myUserName;
	}
	else
	{
		user = serverSIPUser;
	}

	try {
		// incrementing the sequence number of the request
		requestSeqNbr = requestSeqNbr + 1;
		
		//create From Header
		SipURI fromAddress = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, Constants.STR_SIP_FROMADDRESS);
		Address fromNameAddress = addressFactory.createAddress(fromAddress);
		fromNameAddress.setDisplayName(Constants.STR_SIP_FROMDISPLAYNAME);
		FromHeader fromHeader = headerFactory.createFromHeader(fromNameAddress, String.valueOf(myIdTag));
		// create To Header
		SipURI toAddress = addressFactory.createSipURI(user, serverSIPDomain);
		Address toNameAddress = addressFactory.createAddress(toAddress);
		toNameAddress.setDisplayName(Constants.STR_SIP_TODISPLAYNAME);
		ToHeader toHeader = headerFactory.createToHeader(toNameAddress, null);
		
		// create Request URI
		String addressAndPort = serverIPAddress + ":" + serverPort; 
		SipURI requestURI = addressFactory.createSipURI(user, addressAndPort);
		
		// Create ViaHeaders
		ArrayList viaHeaders = new ArrayList();

		ViaHeader viaHeader = headerFactory.createViaHeader(
				sipStack.getIPAddress(),
				sipProvider.getListeningPoint().getPort(),
				transportType,
				null);
		
		// add via headers
		viaHeaders.add(viaHeader);

		// Create ContentTypeHeader
		ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("application", "sdp");

		// Create a new CallId header
		CallIdHeader callIdHeader = sipProvider.getNewCallId();

		// Create a new MaxForwardsHeader
		MaxForwardsHeader maxForwards = headerFactory.createMaxForwardsHeader(70);

		// setting the primitive to execute
		switch (number) {
		case Constants.CONTROL_REGISTER :	
			strRequest = Request.REGISTER;
			break;
		case Constants.CONTROL_INVITE :
			strRequest = Request.INVITE;
			break;
		case Constants.CONTROL_ACK :
			strRequest = Request.ACK;
			break;
		default :
			strRequest = Request.REGISTER;
		}
		lastControlRequestSent = strRequest;
		// Create a new Cseq header
		CSeqHeader cSeqHeader =
			headerFactory.createCSeqHeader(1, strRequest);
		
		// Create the request.
		Request request = messageFactory.createRequest(
				requestURI,
				strRequest,
				callIdHeader,
				cSeqHeader,
				fromHeader,
				toHeader,
				viaHeaders,
				maxForwards);

		String host = sipStack.getIPAddress();
		
		SipURI contactUrl = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, host);
		contactUrl.setPort(listeningPoint.getPort());
		
		// Create the contact name address.
		SipURI contactURI = addressFactory.createSipURI(Constants.STR_SIP_FROMNAME, host);
		contactURI.setPort(sipProvider.getListeningPoint().getPort());

		Address contactAddress = addressFactory.createAddress(contactURI);

		// Add the contact address.
		contactAddress.setDisplayName(Constants.STR_SIP_FROMNAME);

		contactHeader = headerFactory.createContactHeader(contactAddress);
		request.addHeader(contactHeader);

		// Add the extension header for the ID SessionObject.
		Header extensionHeader =
			headerFactory.createHeader(Constants.STR_HEADERNAME, String.valueOf(myIdTag));
		request.addHeader(extensionHeader);

		//Add the extension header for the sequence ID
		Header extensionHeader2 =
			headerFactory.createHeader(Constants.STR_HEADERCSEQ, String.valueOf(requestSeqNbr));
		request.addHeader(extensionHeader2);
		
		// setting the SDP params
		String sdpData = SdpV
		+ SdpO
		+ SdpS
		+ SdpC
		+ SdpT
		+ SdpM;
	for (int i=0; i<SdpA.size();i++)
		sdpData=sdpData+SdpA.get(i);
		
		request.setContent(sdpData, contentTypeHeader);


		Header callInfoHeader =
			headerFactory.createHeader(
				"Call-Info",
				"<http://www.antd.nist.gov>");
		request.addHeader(callInfoHeader);

		firstTID = sipProvider.getNewClientTransaction(request);
		
		dateInvite= System.currentTimeMillis();

		firstTID.sendRequest();
		
		// sets the response code to 100 ... 
		// we need this be sure to capture all the INVITE responses (many 100's)
		// does not influence the other requests
		lastResponseCode = 100;
	}catch(Exception e) {
		throw new IsacRuntimeException("DoRequest failed : " + e);
	}
	
}

/**
 * Waits for a request to come by doing a P on the semaphore.
 * puts a flag to true...a request will be received only if this flag is up
 * (see processRequest)
 * 
 */
public void doWaitRequest()
{
	try {	
		// This object is waiting for a request to come
		isWaitingForRequest = true;
		Semaphore so = (Semaphore)myMaster.listResponseCode.get((myIdTag-1)); 
		so.P();
		
		// The object received a request or exited with a timeout...is not waiting any more
		isWaitingForRequest = false;
		lastRequestCode = so.getValeur();
		
	}catch(Exception e) {
		throw new IsacRuntimeException("DoRequest failed : " + e);
	}
	
}	

/**
 * This method will send a response Code to the system under
 * test. It is a control action so it does not wait for anything
 * and returns void
 *
 * @param params :
 *            A map of the parameters containing the code to be sent
 * 
 */
private void doControlRespond(Map params) {
		// Testing if a transaction is initialized and a request has been received
		if ((ServerTID == null) || (lastRequest == null))
			return;
		
		// getting the code to send from the params
		String strCodeToRespond = (String) params.get(Constants.STR_CONTROL_PARAM_CODETORESPOND);
		int intCodeToRespond = Integer.parseInt(strCodeToRespond);
		
		try {
			// creating and sending the response
			Response response = messageFactory.createResponse(intCodeToRespond,lastRequest);
			ServerTID.sendResponse(response);
		} catch (Exception e) {
			throw new IsacRuntimeException("SipStack destruction failed : " + e);
			
		}	
	}
/***************************************************************************
 * TOOLS*
 **************************************************************************/
/**
 * Auxiliairy tool method used to parse a heder of type : HEADER_NAME:ID_VALUE
 * and to extract the ID_VALUE of it.
 * 
 * @param 
 * 		header : the header to be parsed
 * @return Returns the parsed ID_VALUE string 
 */
private String getTagFromIdHeader(String header) {
	String[] headerStrings = header.split(": ");
	headerStrings = headerStrings[1].split("\r\n");
	return headerStrings[0];
}


		/***********************************************************************
		 * GETTERS AND SETTERS *
		 **********************************************************************/

	/**
	 * @return Returns the addressFactory.
	 */
	public AddressFactory getAddressFactory() {
		return addressFactory;
	}

	/**
	 * @param addressFactory
	 *            The addressFactory to set.
	 */
	public void setAddressFactory(AddressFactory addressFactory) {
		this.addressFactory = addressFactory;
	}

	/**
	 * @return Returns the contactHeader.
	 */
	public ContactHeader getContactHeader() {
		return contactHeader;
	}

	/**
	 * @param contactHeader
	 *            The contactHeader to set.
	 */
	public void setContactHeader(ContactHeader contactHeader) {
		this.contactHeader = contactHeader;
	}

	/**
	 * @return Returns the headerFactory.
	 */
	public HeaderFactory getHeaderFactory() {
		return headerFactory;
	}

	/**
	 * @param headerFactory
	 *            The headerFactory to set.
	 */
	public void setHeaderFactory(HeaderFactory headerFactory) {
		this.headerFactory = headerFactory;
	}

	/**
	 * @return Returns the listeningPoint.
	 */
	public ListeningPoint getListeningPoint() {
		return listeningPoint;
	}

	/**
	 * @param listeningPoint
	 *            The listeningPoint to set.
	 */
	public void setListeningPoint(ListeningPoint listeningPoint) {
		this.listeningPoint = listeningPoint;
	}

	/**
	 * @return Returns the myIPAdress.
	 */
	public String getMyIPAddress() {
		return myIPAddress;
	}

	/**
	 * @param myIPAdress
	 *            The myIPAdress to set.
	 */
	public void setMyIPAddress(String myIPAdress) {
		this.myIPAddress = myIPAdress;
	}

	/**
	 * @return Returns the serverIPAddress.
	 */
	public String getServerIPAddress() {
		return serverIPAddress;
	}

	/**
	 * @param serverIPAddress
	 *            The registrarIPAddress to set.
	 */
	public void setServerIPAddress(String IPAddress) {
		serverIPAddress = IPAddress;
	}

	/**
	 * @return Returns the sipProvider.
	 */
	public SipProvider getSipProvider() {
		return sipProvider;
	}

	/**
	 * @param sipProvider
	 *            The sipProvider to set.
	 */
	public void setSipProvider(SipProvider sipProvider) {
		this.sipProvider = sipProvider;
	}

	/**
	 * @return Returns the sipStack.
	 */
	public SipStack getSipStack() {
		return sipStack;
	}

	/**
	 * @param sipStack
	 *            The sipStack to set.
	 */
	public void setSipStack(SipStack sipStack) {
		this.sipStack = sipStack;
	}

	/**
	 * @return Returns the transportType.
	 */
	public String getTransportType() {
		return transportType;
	}

	/**
	 * @param transportType
	 *            The transportType to set.
	 */
	public void setTransportType(String transportType) {
		this.transportType = transportType;
	}

	public int getLastResponseCode() {
		return lastResponseCode;
	}

	public void setLastResponseCode(int lastResponseCode) {
		this.lastResponseCode = lastResponseCode;
	}

	/**
	 * @return Returns the messageFactory.
	 */
	public MessageFactory getMessageFactory() {
		return messageFactory;
	}

	/**
	 * @param messageFactory
	 *            The messageFactory to set.
	 */
	public void setMessageFactory(MessageFactory messageFactory) {
		this.messageFactory = messageFactory;
	}

	/**
	 * @return Returns the serverSIPDomain.
	 */
	public String getServerSIPDomain() {
		return serverSIPDomain;
	}

	/**
	 * @param serverSIPDomain
	 *            The serverSIPDomain to set.
	 */
	public void setServerSIPDomain(String serverSIPDomain) {
		this.serverSIPDomain = serverSIPDomain;
	}

	/**
	 * @return Returns the serverSIPUser.
	 */
	public String getServerSIPUser() {
		return serverSIPUser;
	}

	/**
	 * @param serverSIPUser
	 *            The serverSIPUser to set.
	 */
	public void setServerSIPUser(String serverSIPUser) {
		this.serverSIPUser = serverSIPUser;
	}

	/**
	 * @return Returns the dialog.
	 */
	public Dialog getDialog() {
		return dialog;
	}

	/**
	 * @param dialog
	 *            The dialog to set.
	 */
	public void setDialog(Dialog dialog) {
		this.dialog = dialog;
	}

	public String getMyUserName() {
		return myUserName;
	}

	public void setMyUserName(String myUserName) {
		this.myUserName = myUserName;
	}

	/**
	 * @return Returns the dateInvite.
	 */
	public long getDateInvite() {
		return dateInvite;
	}

	/**
	 * @param dateInvite The dateInvite to set.
	 */
	public void setDateInvite(long dateInvite) {
		this.dateInvite = dateInvite;
	}

	/**
	 * @return Returns the lastRequest.
	 */
	public Request getLastRequest() {
		return lastRequest;
	}

	/**
	 * @param lastRequest The lastRequest to set.
	 */
	public void setLastRequest(Request lastRequest) {
		this.lastRequest = lastRequest;
	}

	/**
	 * @return Returns the timeout.
	 */
	public long getTimeout() {
		return timeOut;
	}

	/**
	 * @param timeout The timeout to set.
	 */
	public void setTimeout(long timeout) {
		this.timeOut = timeout;
	}

	/**
	 * @return Returns the myPort.
	 */
	public int getMyPort() {
		return myPort;
	}

	/**
	 * @param myPort The myPort to set.
	 */
	public void setMyPort(int myPort) {
		this.myPort = myPort;
	}

	/**
	 * @return Returns the serverPort.
	 */
	public String getServerPort() {
		return serverPort;
	}

	/**
	 * @param serverPort The serverPort to set.
	 */
	public void setServerPort(String serverPort) {
		this.serverPort = serverPort;
	}

}
